#include <p32mx/p32mx_core.h>
#include <p32mx/interrupt.h>
#include <stdint.h>
#include <string.h>

extern int (* __vectors[64])(void);

void interrupt_enable()
{
	uint32_t c0_status;
	uint32_t c0_cause;
	int i;

	c0_status = read_c0_status();
	c0_status &= ~0x1;		/* IE=0 */
	write_c0_status(c0_status);
#if defined(__32M4KCORE__)
{
	uint32_t srsctl,hss;
	srsctl = read_c0_srsctl();
	hss = (srsctl >> C0_SRSCTL_HSS_BIT ) & 0xf;
	srsctl = ( hss << C0_SRSCTL_EICSS_BIT );
	write_c0_srsctl(srsctl);
}
#endif
#if defined(__32M4KCORE__) && defined(USE_SRS)
	INTCON_SET = (1 << INTCON_SS0);
#elif defined(__32M4KCORE__)
	INTCON_CLR = (1 << INTCON_SS0);
#endif
	c0_cause = read_c0_cause();
	write_c0_cause(c0_cause | C0_CAUSE_IV); /* USE External INTC */
	INTCON_CLR = (1 << INTCON_MVEC);	/* Single Vector Mode */

	c0_status = read_c0_status();
	c0_status &= 0xffa70000;	/* BEV=0 , SR=0,  NMI=0 */
					/* UM=0 , ERL=0 , EXL=0, IE=0 */
	c0_status |= 0x0001;		/* IPL=0 , IE = 1 */

	//INTSTAT = (1 << INTSTAT_RIPL);	/* Req IPL=1 , VEC=0 ?? */
	for (i=0; i<10; i++) {
		IPC(i) = 0x04040404;	/* IPxx = 1, ISxx = 0 ?? */
	}
	/* unmask all first */
	IEC_CLR(0) = 0xffffffff;
	IEC_CLR(1) = 0xffffffff;
	/* clear all interrupts */
	IFS_CLR(0) = 0xffffffff;
	IFS_CLR(1) = 0xffffffff;
	write_c0_status(c0_status);
}

#if defined(__32M4KCORE__)
#define ffs(x)	__fls(x)
#endif

extern void __intent();

uint32_t intent(uint32_t c0_status, uint32_t c0_cause, uint32_t *c0_epc_p)
{
	uint32_t isr;

#if defined(__32M4KCORE__) && defined(USE_SRS)
#else
	if (c0_epc_p && ((uint32_t)__intent <= *c0_epc_p )
		     && (*c0_epc_p < (uint32_t)__intent + 0x200 )) {
		*c0_epc_p = read_k1();
	}
#endif
	write_c0_cause(c0_cause & (C0_CAUSE_DC|C0_CAUSE_IV));
	/* kernel mode + irq_disable */
	write_c0_status(c0_status & ~(C0_STATUS_UM | C0_STATUS_ERL
				    | C0_STATUS_EXL| C0_STATUS_IE ));

	while ((isr = (IFS(0) & IEC(0)))) {
		int r;
		isr = ffs(isr) -1;
		__intc_mask_irq(isr);
		r = (__vectors[isr])();
		if (r) {
			__intc_unmask_irq(isr);
		}
	}
	while ((isr = (IFS(1) & IEC(1)))) {
		int r;
		isr = ffs(isr) -1;
		__intc_mask_irq(isr | 0x20);
		r = (__vectors[isr | 0x20])();
		if (r) {
			__intc_unmask_irq(isr | 0x20);
		}
	}
#if defined(__32M4KCORE__) && defined(USE_SRS)
	c0_status &= ~( C0_STATUS_IPL_MASK);	/* clear IPL only (?) */
#else
	/* clear IPL, ERL,EXL */
	c0_status &= ~( C0_STATUS_IPL_MASK | C0_STATUS_ERL | C0_STATUS_EXL);
#endif
	return c0_status;
}
