#include <p32mx/p32mx_intc.h>
#include <p32mx/p32mx_timer.h>
#include <p32mx/p32mx_clkc.h>
#include <p32mx/p32mx_core.h>
#include <p32mx/interrupt.h>
#include <p32mx/ostimer.h>

#define JZ_OSTIMER_LO     3       /* TIMER4 */
#define JZ_OSTIMER_HI     4       /* TIMER5 */
#define JZ_OSTIMER_IRQ    IRQ_T5

#define OSTIMER_DEBUG

static volatile uint32_t current_cycle_high = 0;

union cycle_type {
	uint64_t cycle64;
	uint32_t cycle32[2];
};

uint32_t ostimer32()
{
	uint16_t hi,lo1,lo2;
	lo1 = TMR(JZ_OSTIMER_LO);
	hi = TMR(JZ_OSTIMER_HI);
	lo2 = TMR(JZ_OSTIMER_LO);
	if (lo1 < lo2) return ((uint32_t)hi << 16 ) | lo2;
	return ((uint32_t)TMR(JZ_OSTIMER_HI) << 16 ) | lo2;
}

uint64_t ostimer64()
{
	uint32_t hi,lo1,lo2;
	lo1 = ostimer32();
	hi = current_cycle_high;
	lo2  = ostimer32();
	if (lo1 < lo2) return ((uint64_t)hi << 32 ) | lo2;
	return ((uint64_t)current_cycle_high << 32 ) | lo2;
}

void (* ostimer_callback)(void) __attribute__((weak));

SIGNAL(JZ_OSTIMER_IRQ)
{
	__intc_ack_irq(JZ_OSTIMER_IRQ);
	current_cycle_high++;
	if (ostimer_callback) (ostimer_callback)();
	return IRQ_HANDLED;
}

int ostimer_clk()
{
	int div;
	
	div = (T_CON(JZ_OSTIMER_LO) >> T_CON_TCKPS) & 0x3;
	if (div == 7) div++;
	div +=  ( OSCCON >> OSCCON_PBDIV ) & 0x3;
	return clkc_get_sysclk() >> div;
}

void __ostimer_start(uint32_t prescale)
{
	__intc_mask_irq(JZ_OSTIMER_IRQ);
	current_cycle_high = 0;
	/* disable timer clock */
	T_CON(JZ_OSTIMER_HI) = 0;
	T_CON(JZ_OSTIMER_LO) = ( prescale & 0x3) << T_CON_TCKPS;
	__intc_ack_irq(JZ_OSTIMER_IRQ);
	TMR(JZ_OSTIMER_LO) = 0;
	TMR(JZ_OSTIMER_HI) = 0;
#ifdef OSTIMER_DEBUG
	/* free running */
	PR(JZ_OSTIMER_LO) = 0xffff;
	PR(JZ_OSTIMER_HI) = 0x0fff;
#else
	/* free running */
	PR(JZ_OSTIMER_LO) = 0xffff;
	PR(JZ_OSTIMER_HI) = 0xffff;
#endif
	/* start with 32bit mode */
	T_CON_SET(JZ_OSTIMER_LO) = (1 << T_CON_T32) | (1 << T_CON_ON); 
	__intc_unmask_irq(JZ_OSTIMER_IRQ);
}

void ostimer_start(int is_fast)
{
	uint32_t prescale;
	if (!is_fast) {
		prescale = 6; // 1/64
	} else {
		prescale = 0; // 1/1
	}
	__ostimer_start(prescale);
}


void ostimer_stop()
{
	__intc_mask_irq(JZ_OSTIMER_IRQ);
	/* disable timer clock */
	T_CON(JZ_OSTIMER_LO) = 0;
	T_CON(JZ_OSTIMER_HI) = 0;
	__intc_ack_irq(JZ_OSTIMER_IRQ);
}
